from typing import Optional, Tuple, Union

import matplotlib.pyplot as plt
from numpy import typing as npt
import numpy as np

from eyepy.core.annotations import EyeEnfacePixelAnnotation


class EyeEnface:
    """ """

    def __init__(self, data, meta):
        """

        Args:
            data:
            meta:
        """
        self.data = data
        self._area_maps = []
        self.meta = meta

    @property
    def area_maps(self):
        """

        Returns:

        """
        # Create a dict to access area_maps by their name
        return {am.name: am for am in self._area_maps}

    def add_area_annotation(self, area_map=None, meta=None, **kwargs):
        """

        Args:
            area_map:
            meta:
            **kwargs:

        Returns:

        """
        if meta is None:
            meta = {}
        meta.update(**kwargs)
        area_annotation = EyeEnfacePixelAnnotation(self, area_map, meta)
        self._area_maps.append(area_annotation)
        return area_annotation

    @property
    def scale_x(self):
        """

        Returns:

        """
        return self.meta["scale_x"]

    @property
    def scale_y(self):
        """

        Returns:

        """
        return self.meta["scale_y"]

    @property
    def size_x(self):
        """

        Returns:

        """
        return self.shape[1]

    @property
    def size_y(self):
        """

        Returns:

        """
        return self.shape[0]

    @property
    def laterality(self):
        """

        Returns:

        """
        return self.meta["laterality"]

    @property
    def shape(self):
        """

        Returns:

        """
        return self.data.shape

    def plot(self,
             ax: Optional[plt.Axes] = None,
             region: Union[slice, Tuple[slice, slice]] = np.s_[:, :]):
        """

        Args:
            ax:
            region:

        Returns:

        """
        if ax is None:
            ax = plt.gca()
        ax.imshow(self.data[region], cmap="gray")

        # Make sure tick labels match the image region
        y_start = region[0].start if region[0].start is not None else 0
        x_start = region[1].start if region[1].start is not None else 0
        y_end = region[0].stop if region[0].stop is not None else self.size_y
        x_end = region[1].stop if region[1].stop is not None else self.size_x

        # Ticks are not clipped to the image region. Clip them here.
        yticks = ax.get_yticks()
        yticks = yticks[np.nonzero(
            np.logical_and(yticks >= 0, yticks <= y_end - y_start - 1))]
        xticks = ax.get_xticks()
        xticks = xticks[np.nonzero(
            np.logical_and(xticks >= 0, xticks <= x_end - x_start - 1))]

        # Set clipped ticks (this is only necessary because we change the labels later)
        ax.set_yticks(yticks)
        ax.set_xticks(xticks)

        # Set labels to ticks + start of the region as an offset
        ax.set_yticklabels([str(int(t + y_start)) for t in yticks])
        ax.set_xticklabels([str(int(t + x_start)) for t in xticks])



from collections import defaultdict

# Plotting config

# Line Style for Layers in B-Scan
layer_kwargs = {"linewidth": 1, "linestyle": "-"}
area_kwargs = {"alpha": 0.5}
ascan_kwargs = {"alpha": 0.5}

# Line Style for B-Scan positions on Slo
line_kwargs = {"linewidth": 0.2, "linestyle": "-", "color": "limegreen"}

# Colors for different Layers
_layer_colors = {
    "BM": "F77189",
    "RPE": "97A431",
    "PR1": "36ADA4",
    "EZ": "36ADA4",
    "PR2": "A48CF4",
    "IZ": "A48CF4",
    "ELM": "E68332",
    "ONL": "50B131",
    "OPL": "38AABF",
    "INL": "E866F4",
    "IPL": "BB9832",
    "GCL": "34AF84",
    "RNFL": "3BA3EC",
    "NFL": "3BA3EC",
    "ILM": "F668C2",
}

layer_colors = defaultdict(lambda: "FF0000", **_layer_colors)



"""Inspired by:

https://github.com/ayl/heyexReader/blob/master/heyexReader/volReader.py
https://github.com/FabianRathke/octSegmentation/blob/master/collector/HDEVolImporter.m
"""

import datetime
import logging
from typing import List
import warnings

import construct as cs
import numpy as np
import skimage
from skimage import transform

from eyepy import __version__
from eyepy import EyeBscanMeta
from eyepy import EyeEnface
from eyepy import EyeEnfaceMeta
from eyepy import EyeVolume
from eyepy import EyeVolumeMeta
from eyepy.core.eyebscan import EyeBscan
from eyepy.core.utils import to_vol_intensity
from eyepy.io.utils import _compute_localizer_oct_transform
from eyepy.io.utils import Bscan
from eyepy.io.utils import FloatDate
from eyepy.io.utils import IntDate
from eyepy.io.utils import Localizer
from eyepy.io.utils import Segmentations

logger = logging.getLogger("eyepy.io.HeVolReader")

bscan_format = cs.Struct(
    "version" /
    cs.Default(cs.PaddedString(12, "ascii"), f"ep-v{__version__}".ljust(12)) *
    "Version identifier",
    "bscan_hdr_size" / cs.Int32ul *
    "Size of the B-Scan header in bytes. It is identical to the same value of the file header.",
    "start_x" / cs.Float64l *
    "X-Coordinate of the B-Scan's start point in mm.",
    "start_y" / cs.Float64l *
    "Y-Coordinate of the B-Scan's start point in mm.",
    "end_x" / cs.Float64l *
    "X-Coordinate of the B-Scan's end point in mm. For circle scans, this is the X-Coordinate of the circle's center point.",
    "end_y" / cs.Float64l *
    "Y-Coordinate of the B-Scan's end point in mm. For circle scans, this is the Y-Coordinate of the circle's center point.",
    "num_seg" / cs.Int32ul * "Number of segmentation vectors",
    "seg_offset" / cs.Default(cs.Int32ul, 256) *
    "Offset of the array of segmentation vectors relative to the beginning of this B-Scan header.",
    "quality" / cs.Float32l *
    "Image quality measure. If this value does not exist, its value is set to INVALID.",
    "shift" / cs.Int32ul *
    "Horizontal shift (in # of A-Scans) of the classification band against the segmentation lines (for circular scan only).",
    "iv_transformation" / cs.Float32l[6] *
    "Intra volume transformation matrix. The values are only available for volume and radial scans and if alignment is turned off, otherwise the values are initialized to 0.",
    "__empty" / cs.Padding(168) * "Spare bytes for future use.",
    # cs.Probe(cs.this.bscan_hdr.sizeof(cs.this)),
    "layer_segmentations" / Segmentations *
    "Layer segmentations for 17 Layers. If layers are not segmented the data is empty",
    "__empty" /
    cs.Padding(cs.this.bscan_hdr_size - 256 - 17 * 4 * cs.this._.size_x) *
    "B-scan data starts after 'bscan_hdr_size' bytes from the start of the"
    "B-scan. This header contains 256 bytes for general information, then 17 "
    "layer segmentations each composed size_x * 4 bytes. Then we have padding"
    "to fill the bscan_hdr_size.",
    "data" / Bscan * "B-scan data",
)

vol_format = cs.Struct(
    "version" /
    cs.Default(cs.PaddedString(12, "ascii"), f"ep-v{__version__}".ljust(12)) *
    "Version identifier: HSF-OCT-xxx",
    "size_x" / cs.Int32ul *
    "Number of A-Scans in each B-Scan, i.e. B-Scan width in pixel",
    "n_bscans" / cs.Int32ul * "Number of B-Scans in OCT scan",
    "size_y" / cs.Int32ul *
    "Number of samples in an A-Scan, i.e. B-Scan height in pixel",
    "scale_x" / cs.Float64l * "Width of a B-Scan pixel in mm",
    "distance" / cs.Float64l * "Distance between two adjacent B-Scans in mm",
    "scale_y" / cs.Float64l * "Height of a B-Scan pixel in mm",
    "size_x_slo" / cs.Int32ul * "Width of the SLO image in pixel",
    "size_y_slo" / cs.Int32ul * "Height of the SLO image in pixel",
    "scale_x_slo" / cs.Float64l * "Width of a pixel in the SLO image in mm",
    "scale_y_slo" / cs.Float64l * "Height of a pixel in the SLO image in mm",
    "field_size_slo" / cs.Default(cs.Int32ul, 0) *
    "Horizontal field size of the SLO image in dgr",
    "scan_focus" / cs.Default(cs.Float64l, 0) * "Scan focus in dpt",
    "scan_position" / cs.Default(cs.PaddedString(4, "ascii"), "".ljust(4)) *
    "Examined eye (zero terminated string). 'OS': Left eye; 'OD': Right eye",
    "exam_time" /
    cs.Default(IntDate, datetime.datetime(year=2000, month=1, day=1)) *
    "Examination time. The structure holds an unsigned 64-bit date and time value. "
    "It is encoded as 100ns units since beginning of January 1, 1601",
    "scan_pattern" / cs.Default(cs.Int32sl, 0) * """Scan pattern type:
    0 = Unknown pattern,
    1 = Single line scan (one B-Scan only),
    2 = Circular scan (one B-Scan only),
    3 = Volume scan in ART mode,
    4 = Fast volume scan,
    5 = Radial scan (aka. star pattern)""",
    "bscan_hdr_size" / cs.Default(cs.Int32ul, 61440) *
    "Size of the Header preceding each B-Scan in bytes",
    "id" / cs.PaddedString(16, "ascii") *
    "Unique identifier of this OCT-scan (zero terminated string). This is"
    "identical to the number <SerID> that is part of the file name."
    "Format: n[.m] n and m are numbers. The extension .m exists only for "
    "ScanPattern 1 and 2. Examples: 2390, 3433.2",
    "reference_id" / cs.PaddedString(16, "ascii") *
    "Unique identifier of the reference OCT-scan (zero terminated string)."
    "Format: see ID, This ID is only present if the OCT-scan is part of a "
    "progression otherwise this string is empty. For the reference scan of "
    "a progression ID and ReferenceID are identical.",
    "pid" / cs.Default(cs.Int32ul, 0) * "Internal patient ID used by HEYEX.",
    "patient_id" /
    cs.Default(cs.PaddedString(24, "ascii"), "unknown".ljust(24)) *
    "User-defined patient ID (zero terminated string).",
    "dob" / FloatDate * "Patient's date of birth",
    "vid" / cs.Default(cs.Int32ul, 0) * "Internal visit ID used by HEYEX.",
    "visit_id" /
    cs.Default(cs.PaddedString(24, "ascii"), "unknown".ljust(24)) *
    "User-defined visit ID (zero terminated string). This ID can be defined "
    "in the Comment-field of the Diagnosis-tab of the Examination Data "
    "dialog box. The VisitID must be defined in the first row of the "
    "comment field. It has to begin with an '#' and ends with any "
    "white-space character. It can contain up to 23 alphanumeric "
    "characters (excluding the '#')",
    "visit_date" / FloatDate *
    "Date the visit took place. Identical to the date of an examination tab in HEYEX.",
    "grid_type" / cs.Int32ul *
    "Type of grid used to derive thickness data. 0 No thickness data available, >0 Type"
    " of grid used to derive thickness values. Thickness data is only available for "
    "ScanPattern 3 and 4.",
    "grid_offset" / cs.Int32ul *
    "File offset of the thickness data in the file. If GridType is 0, GridOffset is 0.",
    "grid_type1" / cs.Int32ul *
    "Type of 2nd grid used to derive a 2nd set of thickness data.",
    "grid_offset1" / cs.Int32ul *
    "File offset of the 2nd thickness data set in the file.",
    "prog_id" / cs.PaddedString(34, "ascii") *
    "Internal progression ID (zero terminated string). All scans of the "
    "same progression share this ID.",
    "__empty" / cs.Padding(1790) *
    "Spare bytes for future use. Initialized to 0.",
    "localizer" / Localizer * "NIR localizer image",
    "bscans" / bscan_format[cs.this.n_bscans],
)

# PR1 and EZ map to 14 and PR2 and IZ map to 15. Hence both names can be used
# to access the same data
SEG_MAPPING = {
    "ILM": 0,
    "BM": 1,
    "RNFL": 2,
    "GCL": 3,
    "IPL": 4,
    "INL": 5,
    "OPL": 6,
    "ONL": 7,
    "ELM": 8,
    "IOS": 9,
    "OPT": 10,
    "CHO": 11,
    "VIT": 12,
    "ANT": 13,
    "PR1": 14,
    "PR2": 15,
    "RPE": 16,
}


class HeVolReader:

    def __init__(self, path):
        self.path = path
        with open(self.path, "rb") as vol_file:
            self.parsed_file = vol_format.parse_stream(vol_file)

    @property
    def volume(self) -> EyeVolume:
        ## Check if scan pattern is supported by EyeVolume
        if self.parsed_file.scan_pattern == 2:
            msg = f"The EyeVolume object does not support scan pattern 2 (one Circular B-scan)."
            raise ValueError(msg)
        elif self.parsed_file.scan_pattern == 5:
            msg = f"The EyeVolume object does not support scan pattern 5 (Radial scan - star pattern)."
            raise ValueError(msg)

        data = np.stack([bscan.data for bscan in self.parsed_file.bscans],
                        axis=0)
        volume_meta = self.meta
        localizer = self.localizer
        volume = EyeVolume(
            data=data,
            meta=volume_meta,
            localizer=localizer,
            transformation=_compute_localizer_oct_transform(
                volume_meta, localizer.meta, data.shape),
        )

        layer_height_maps = self.layers
        for name, i in SEG_MAPPING.items():
            volume.add_layer_annotation(layer_height_maps[i], name=name)

        return volume

    @property
    def layers(self):
        layers = np.stack(
            [b.layer_segmentations for b in self.parsed_file.bscans], axis=0)
        layers[layers >= 3.0e+38] = np.nan
        # Currently the shape is (n_bscans, n_layers, width). Swap the first two axes
        # to get (n_layers, n_bscans, width)
        return np.swapaxes(layers, 0, 1)

    @property
    def localizer_meta(self) -> EyeEnfaceMeta:
        return EyeEnfaceMeta(
            scale_x=self.parsed_file.scale_x_slo,
            scale_y=self.parsed_file.scale_y_slo,
            scale_unit="mm",
            modality="NIR",
            laterality=self.parsed_file.scan_position,
            field_size=self.parsed_file.field_size_slo,
            scan_focus=self.parsed_file.scan_focus,
            visit_date=self.parsed_file.visit_date,
            exam_time=self.parsed_file.exam_time,
        )

    @property
    def localizer(self) -> EyeEnface:
        localizer = self.parsed_file.localizer
        return EyeEnface(localizer, self.localizer_meta)

    @property
    def bscan_meta(self) -> List[EyeBscanMeta]:
        return [
            EyeBscanMeta(
                quality=b.quality,
                start_pos=(b.start_x, b.start_y),
                end_pos=(b.end_x, b.end_y),
                pos_unit="mm",
            ) for b in self.parsed_file.bscans
        ]

    @property
    def meta(self) -> EyeVolumeMeta:
        bscan_meta = self.bscan_meta
        return EyeVolumeMeta(
            scale_x=self.parsed_file.scale_x,
            scale_y=self.parsed_file.scale_y,
            scale_z=self.parsed_file.distance,
            scale_unit="mm",
            laterality=self.parsed_file.scan_position,
            visit_date=self.parsed_file.visit_date,
            exam_time=self.parsed_file.exam_time,
            bscan_meta=bscan_meta,
            intensity_transform="vol",
        )


class HeVolWriter:

    def __init__(self, volume: EyeVolume):
        self.volume = volume

    def write(self, path):
        with open(path, "wb") as mf:
            mf.write(self.bytes)

    @property
    def bytes(self):
        return vol_format.build(
            dict(size_x=self.volume.size_x,
                 size_y=self.volume.size_y,
                 n_bscans=self.volume.size_z,
                 scale_x=self.volume.scale_x,
                 scale_y=self.volume.scale_y,
                 distance=self.volume.scale_z,
                 size_x_slo=self.volume.localizer.size_x,
                 size_y_slo=self.volume.localizer.size_y,
                 scale_x_slo=self.volume.localizer.scale_x,
                 scale_y_slo=self.volume.localizer.scale_y,
                 field_size_slo=self.volume.localizer.meta["field_size"],
                 scan_focus=self.volume.localizer.meta["scan_focus"],
                 scan_position=self.volume.localizer.meta["laterality"],
                 exam_time=datetime.datetime.combine(
                     self.volume.meta["exam_time"], datetime.time()),
                 scan_pattern=0,
                 id="unknown",
                 reference_id="unknown",
                 pid=self.volume.meta["patient"]["pid"],
                 patient_id=self.volume.meta["patient"]["patient_id"],
                 dob=datetime.datetime.combine(
                     self.volume.meta["patient"]["dob"], datetime.time()),
                 vid=0,
                 visit_id="unknown",
                 visit_date=datetime.datetime.combine(
                     self.volume.meta["visit_date"], datetime.time()),
                 grid_type=0,
                 grid_offset=0,
                 grid_type1=0,
                 grid_offset1=0,
                 prog_id="unknown",
                 localizer=skimage.img_as_ubyte(self.volume.localizer.data),
                 bscans=self._bscan_dicts))

    @property
    def _bscan_dicts(self):
        return [
            dict(bscan_hdr_size=256 + 17 * 4 * bscan.volume.size_x,
                 start_x=bscan.meta["start_pos"][0],
                 start_y=bscan.meta["start_pos"][1],
                 end_x=bscan.meta["end_pos"][0],
                 end_y=bscan.meta["end_pos"][1],
                 num_seg=len(bscan.layers),
                 quality=bscan.meta["quality"],
                 shift=0,
                 iv_transformation=[0, 0, 0, 0, 0, 0],
                 layer_segmentations=self._segmentations_from_bscan(bscan),
                 data=to_vol_intensity(bscan.data.astype(np.float32)))
            for bscan in self.volume
        ]

    def _segmentations_from_bscan(self, bscan: EyeBscan) -> np.ndarray:
        """ """
        segs = np.zeros((17, bscan.volume.size_x), dtype=np.float32)
        for layer in bscan.layers:
            if layer.name in SEG_MAPPING:
                segs[SEG_MAPPING[layer.name]] = layer.data
            else:
                warnings.warn(
                    f"Unknown layer name: {layer.name}. Skipping. \n VOL format only supports layers with the following names: {list(SEG_MAPPING.keys())}"
                )
        return segs



from __future__ import annotations

from collections import defaultdict
import logging
from typing import List, Optional, Tuple, TYPE_CHECKING

from matplotlib import cm
from matplotlib import colors
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import numpy.typing as npt
from skimage import transform

from eyepy import config

if TYPE_CHECKING:
    from eyepy import EyeEnface
    from eyepy import EyeVolume

logger = logging.getLogger("eyepy.core.annotations")


class EyeVolumeLayerAnnotation:
    """ """

    def __init__(
        self,
        volume: EyeVolume,
        data: Optional[npt.NDArray[np.float32]] = None,
        meta: Optional[dict] = None,
        **kwargs,
    ):
        """

        Args:
            volume:
            data:
            meta:
            **kwargs:
        """
        self.volume = volume
        if data is None:
            self.data = np.full((volume.size_z, volume.size_x), np.nan)
        else:
            self.data = data

        if meta is None:
            self.meta = kwargs
        else:
            self.meta = meta
            self.meta.update(**kwargs)

        # knots is a dict layername: list of curves where every curve is a list of knots
        if "knots" not in self.meta:
            self.meta["knots"] = defaultdict(lambda: [])
        elif type(self.meta["knots"]) is dict:
            self.meta["knots"] = defaultdict(lambda: [], self.meta["knots"])

        if "name" not in self.meta:
            self.meta["name"] = "Layer Annotation"

        self.meta["current_color"] = config.layer_colors[self.name]

    @property
    def name(self):
        """

        Returns:

        """
        return self.meta["name"]

    @name.setter
    def name(self, value):
        self.meta["name"] = value

    @property
    def knots(self):
        """

        Returns:

        """
        return self.meta["knots"]

    def layer_indices(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """ Returns pixel indices of the layer in the volume

        While the layer is stored as the offset from the bottom of the OCT volume, some applications require
        layer discretized to voxel positions. This method returns the layer as indices into the OCT volume.

        The indices can be used for example to create layer maps for semantic segmentation.

        ```python
        import matplotlib.pyplot as plt
        import numpy as np
        import eyepy as ep

        eye_volume = ep.data.load("drusen_patient")
        rpe_annotation = eye_volume.layers["RPE"]
        rpe_indices = rpe_annotation.layer_indices()
        rpe_map = np.zeros(eye_volume.shape)
        rpe_map[rpe_indices] = 1
        plt.imshow(rpe_map[0]) # (1)
        ```

        1.  Visualize layer map for the first B-scan

        Returns:
            A tuple with indices for the layers position in the volume - Tuple[bscan_indices, row_indices, column_indices]
        """
        layer = self.data[:, np.newaxis, :]
        nan_indices = np.isnan(layer)
        row_indices = np.rint(layer).astype(int)[~nan_indices]
        x = np.ones(layer.shape)
        x[nan_indices] = 0
        bscan_indices, _, col_indices = np.nonzero(x)
        return (bscan_indices, row_indices, col_indices)


class EyeVolumePixelAnnotation:
    """ """

    def __init__(
            self,
            volume: EyeVolume,
            # Type hint for an optional boolean numpy array
            data: Optional[npt.NDArray[np.bool_]] = None,
            meta: Optional[dict] = None,
            radii=(1.5, 2.5),
            n_sectors=(1, 4),
            offsets=(0, 45),
            center=None,
            **kwargs,
    ):
        """

        Args:
            volume:
            data:
            meta:
            radii:
            n_sectors:
            offsets:
            center:
            **kwargs:
        """
        self.volume = volume

        if data is None:
            self.data = np.full(self.volume.shape,
                                fill_value=False,
                                dtype=bool)
        else:
            self.data = data

        self._masks = None
        self._quantification = None

        if meta is None:
            self.meta = kwargs
        else:
            self.meta = meta
            self.meta.update(**kwargs)

        self.meta.update(
            **{
                "radii": radii,
                "n_sectors": n_sectors,
                "offsets": offsets,
                "center": center,
            })

        if "name" not in self.meta:
            self.meta["name"] = "Voxel Annotation"

    @property
    def name(self):
        """

        Returns:

        """
        return self.meta["name"]

    @name.setter
    def name(self, value):
        self.meta["name"] = value

    def _reset(self):
        self._masks = None
        self._quantification = None

    @property
    def radii(self):
        """

        Returns:

        """
        return self.meta["radii"]

    @radii.setter
    def radii(self, value):
        self._reset()
        self.meta["radii"] = value

    @property
    def n_sectors(self):
        """

        Returns:

        """
        return self.meta["n_sectors"]

    @n_sectors.setter
    def n_sectors(self, value):
        self._reset()
        self.meta["n_sectors"] = value

    @property
    def offsets(self):
        """

        Returns:

        """
        return self.meta["offsets"]

    @offsets.setter
    def offsets(self, value):
        self._reset()
        self.meta["offsets"] = value

    @property
    def center(self):
        """

        Returns:

        """
        return self.meta["center"]

    @center.setter
    def center(self, value):
        self._reset()
        self.meta["center"] = value

    @property
    def projection(self):
        """

        Returns:

        """

        # The flip is required because in the volume the bottom most B-scan has the lowest index
        # while in the enface projection the bottom most position should have the biggest index.
        return np.flip(np.nansum(self.data, axis=1), axis=0)

    @property
    def enface(self):
        """

        Returns:

        """
        return transform.warp(
            self.projection,
            self.volume.localizer_transform.inverse,
            output_shape=(
                self.volume.localizer.size_y,
                self.volume.localizer.size_x,
            ),
            order=0,
        )

    def plot(
        self,
        ax=None,
        region=np.s_[...],
        cmap="Reds",
        vmin=None,
        vmax=None,
        cbar=True,
        alpha=1,
    ):
        """

        Args:
            ax:
            region:
            cmap:
            vmin:
            vmax:
            cbar:
            alpha:

        Returns:

        """
        enface_projection = self.enface

        if ax is None:
            ax = plt.gca()

        if vmin is None:
            vmin = 1
        if vmax is None:
            vmax = max([enface_projection.max(), vmin])

        enface_crop = enface_projection[region]
        visible = np.zeros(enface_crop.shape)
        visible[np.logical_and(vmin <= enface_crop, enface_crop <= vmax)] = 1

        if cbar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            plt.colorbar(
                cm.ScalarMappable(colors.Normalize(vmin=vmin, vmax=vmax),
                                  cmap=cmap),
                cax=cax,
            )

        ax.imshow(
            enface_crop,
            alpha=visible * alpha,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )

    @property
    def masks(self):
        """

        Returns:

        """
        from eyepy.core.grids import grid

        if self._masks is None:
            self._masks = grid(
                mask_shape=self.volume.localizer.shape,
                radii=self.radii,
                laterality=self.volume.laterality,
                n_sectors=self.n_sectors,
                offsets=self.offsets,
                radii_scale=self.volume.scale_x,
                center=self.center,
            )

        return self._masks

    @property
    def quantification(self):
        """

        Returns:

        """
        if self._quantification is None:
            self._quantification = self._quantify()

        return self._quantification

    def _quantify(self):
        enface_voxel_size_ym3 = (self.volume.localizer.scale_x * 1e3 *
                                 self.volume.localizer.scale_y * 1e3 *
                                 self.volume.scale_y * 1e3)
        oct_voxel_size_ym3 = (self.volume.scale_x * 1e3 * self.volume.scale_z *
                              1e3 * self.volume.scale_y * 1e3)

        enface_projection = self.enface

        results = {}
        for name, mask in self.masks.items():
            results[f"{name} [mm³]"] = ((enface_projection * mask).sum() *
                                        enface_voxel_size_ym3 / 1e9)

        results["Total [mm³]"] = enface_projection.sum(
        ) * enface_voxel_size_ym3 / 1e9
        results["Total [OCT voxels]"] = self.projection.sum()
        results["OCT Voxel Size [µm³]"] = oct_voxel_size_ym3
        results["Laterality"] = self.volume.laterality
        return results

    # # Todo
    # def create_region_shape_primitives(
    #     mask_shape,
    #     radii: list = (0.8, 1.8),
    #     n_sectors: list = (1, 4),
    #     rotation: list = (0, 45),
    #     center=None,
    # ):
    #     """Create circles and lines indicating region boundaries of quantification
    #     masks. These can be used for plotting the masks.
    #
    #     Parameters
    #     ----------
    #     mask_shape :
    #     radii :
    #     n_sectors :
    #     rotation :
    #     center :
    #
    #     Returns
    #     -------
    #     """
    #     if center is None:
    #         center = (mask_shape[0] / 2, mask_shape[0] / 2)
    #
    #     primitives = {"circles": [], "lines": []}
    #     # Create circles
    #     for radius in radii:
    #         primitives["circles"].append({"center": center, "radius": radius})
    #
    #     for i, (n_sec, rot, radius) in enumerate(zip(n_sectors, rotation, radii)):
    #         rot = rot / 360 * 2 * np.pi
    #         if not n_sec is None and n_sec != 1:
    #             for sec in range(n_sec):
    #                 theta = 2 * np.pi / n_sec * sec + rot
    #
    #                 start = cmath.rect(radii[i - 1], theta)
    #                 start = (start.real + center[0], start.imag + center[1])
    #
    #                 end = cmath.rect(radius, theta)
    #                 end = (end.real + center[0], end.imag + center[1])
    #
    #                 primitives["lines"].append({"start": start, "end": end})
    #
    #     return primitives

    def plot_quantification(
        self,
        ax=None,
        region=np.s_[...],
        alpha=0.5,
        vmin=None,
        vmax=None,
        cbar=True,
        cmap="YlOrRd",
    ):
        """

        Args:
            ax:
            region:
            alpha:
            vmin:
            vmax:
            cbar:
            cmap:

        Returns:

        """

        if ax is None:
            ax = plt.gca()

        mask_img = np.zeros(self.volume.localizer.shape, dtype=float)[region]
        visible = np.zeros_like(mask_img)
        for mask_name in self.masks.keys():
            mask_img += (self.masks[mask_name][region] *
                         self.quantification[mask_name + " [mm³]"])
            visible += self.masks[mask_name][region]

        if vmin is None:
            vmin = mask_img[visible.astype(int)].min()
        if vmax is None:
            vmax = max([mask_img.max(), vmin])

        if cbar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            plt.colorbar(
                cm.ScalarMappable(colors.Normalize(vmin=vmin, vmax=vmax),
                                  cmap=cmap),
                cax=cax,
            )

        ax.imshow(
            mask_img,
            alpha=visible * alpha,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )


class EyeBscanLayerAnnotation:
    """ """

    def __init__(self, eyevolumelayerannotation: EyeVolumeLayerAnnotation,
                 index: int) -> None:
        """

        Args:
            eyevolumelayerannotation:
            index:
        """
        self.eyevolumelayerannotation = eyevolumelayerannotation
        self.volume = eyevolumelayerannotation.volume
        self.index = index

    @property
    def name(self):
        """

        Returns:

        """
        return self.eyevolumelayerannotation.meta["name"]

    @name.setter
    def name(self, value: str):
        self.eyevolumelayerannotation.meta["name"] = value

    @property
    def data(self):
        """

        Returns:

        """
        return self.eyevolumelayerannotation.data[self.index, :]

    @data.setter
    def data(self, value):
        self.eyevolumelayerannotation.data[self.index, :] = value

    @property
    def knots(self):
        """

        Returns:

        """
        return self.eyevolumelayerannotation.knots[self.index]

    @knots.setter
    def knots(self, value: List):
        self.eyevolumelayerannotation.knots[self.index] = value


class EyeEnfacePixelAnnotation:
    """ """

    def __init__(
        self,
        enface: EyeEnface,
        data: Optional[npt.NDArray[np.bool_]] = None,
        meta: Optional[dict] = None,
        **kwargs,
    ):
        """

        Args:
            enface:
            data:
            meta:
            **kwargs:
        """
        self.enface = enface

        if data is None:
            self.data = np.full(self.enface.shape,
                                fill_value=False,
                                dtype=bool)
        else:
            self.data = data

        if meta is None:
            self.meta = kwargs
        else:
            self.meta = meta
            self.meta.update(**kwargs)

        if "name" not in self.meta:
            self.meta["name"] = "Pixel Annotation"

    @property
    def name(self):
        """

        Returns:

        """
        return self.meta["name"]

    @name.setter
    def name(self, value):
        self.meta["name"] = value



"""

The eyepy top

+ [EyeVolume][eyepy.core.eyevolume.EyeVolume]
+ [EyeBscan][eyepy.core.eyebscan.EyeBscan]
+ [EyeEnface][eyepy.core.eyeenface.EyeEnface]

+ [EyeBscanMeta][eyepy.core.eyemeta.EyeBscanMeta]
+ [EyeEnfaceMeta][eyepy.core.eyemeta.EyeEnfaceMeta]
+ [EyeVolumeMeta][eyepy.core.eyemeta.EyeVolumeMeta]

+ [EyeVolumeLayerAnnotation][eyepy.core.annotations.EyeVolumeLayerAnnotation]
+ [EyeBscanLayerAnnotation][eyepy.core.annotations.EyeBscanLayerAnnotation]

+ [EyeEnfacePixelAnnotation][eyepy.core.annotations.EyeEnfacePixelAnnotation]

+ [EyeVolumePixelAnnotation][eyepy.core.annotations.EyeVolumePixelAnnotation]

"""
# isort: skip_file

__author__ = """Olivier Morelle"""
__email__ = "oli4morelle@gmail.com"
__version__ = "0.9.0"

from eyepy.core import drusen
from eyepy.core import EyeBscan
from eyepy.core import EyeBscanLayerAnnotation
from eyepy.core import EyeBscanMeta
from eyepy.core import EyeEnface
from eyepy.core import EyeEnfaceMeta
from eyepy.core import EyeEnfacePixelAnnotation
from eyepy.core import EyeVolume
from eyepy.core import EyeVolumeLayerAnnotation
from eyepy.core import EyeVolumeMeta
from eyepy.core import EyeVolumePixelAnnotation
from eyepy.io import import_bscan_folder
from eyepy.io import import_duke_mat
from eyepy.io import import_heyex_e2e
from eyepy.io import import_heyex_vol
from eyepy.io import import_heyex_xml
from eyepy.io import import_retouch
from eyepy import data



import datetime
import json
import os
from typing import List, MutableMapping, Tuple


class EyeMeta(MutableMapping):
    """ """

    def __init__(self, *args, **kwargs):
        """

        Args:
            *args:
            **kwargs:
        """
        self._store = dict()
        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def as_dict(self):
        """

        Returns:

        """
        data = self._store.copy()
        for key in ["visit_date", "exam_time"]:
            if key in data.keys() and data[key] is not None:
                data[key] = data[key].isoformat()
        return data

    def __getitem__(self, key):
        return self._store[key]

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __str__(self):
        return f"{os.linesep}".join([f"{f}: {self[f]}" for f in self if f != "__empty"])

    def __repr__(self):
        return self.__str__()


class EyeEnfaceMeta(EyeMeta):
    """ """

    def __init__(self, scale_x: float, scale_y: float, scale_unit: str, **kwargs):
        """A dict with required keys to hold meta data for enface images of the eye

        Args:
            scale_x: Horizontal scale of the enface pixels
            scale_y: Vertical scale of the enface pixels
            scale_unit: Unit of the scale. e.g. µm if scale is given in µm/pixel
            **kwargs:
        """
        super().__init__(
            scale_x=scale_x, scale_y=scale_y, scale_unit=scale_unit, **kwargs
        )

    @classmethod
    def from_dict(cls, data: dict):
        """

        Args:
            data:

        Returns:

        """
        for key in ["visit_date", "exam_time"]:
            if key in data.keys() and data[key] is not None:
                data[key] = datetime.datetime.fromisoformat(data[key])
        return cls(**data)


class EyeBscanMeta(EyeMeta):
    """ """

    def __init__(
        self,
        start_pos: Tuple[float, float],
        end_pos: Tuple[float, float],
        pos_unit: str,
        **kwargs,
    ):
        """A dict with required keys to hold meta data for OCT B-scans

        Args:
            start_pos: B-scan start on the enface (in enface space)
            end_pos: B-scan end on the enface (in enface space)
            pos_unit: Unit of the positions
            **kwargs:
        """
        super().__init__(
            start_pos=start_pos, end_pos=end_pos, pos_unit=pos_unit, **kwargs
        )


class EyeVolumeMeta(EyeMeta):
    """ """

    def __init__(
        self,
        scale_z: float,
        scale_x: float,
        scale_y: float,
        scale_unit: str,
        bscan_meta: List[EyeBscanMeta],
        **kwargs,
    ):
        """A dict with required keys to hold meta data for OCT volumes

        Args:
            scale_z: Distance between neighbouring B-scans
            scale_x: Horizontal scale of the B-scan pixels
            scale_y: Vertical scale of the B-scan pixels
            scale_unit: Unit of the scale. e.g. µm if scale is given in µm/pixel
            bscan_meta: A list holding an EyeBscanMeta object for every B-scan of the volume
            **kwargs:
        """
        super().__init__(
            scale_z=scale_z,
            scale_x=scale_x,
            scale_y=scale_y,
            scale_unit=scale_unit,
            bscan_meta=bscan_meta,
            **kwargs,
        )

    def as_dict(self):
        """

        Returns:

        """
        data = super().as_dict()
        data["bscan_meta"] = [bm.as_dict() for bm in data["bscan_meta"]]
        return data

    @classmethod
    def from_dict(cls, data: dict):
        """

        Args:
            data:

        Returns:

        """
        data["bscan_meta"] = [EyeBscanMeta(**d) for d in data["bscan_meta"]]
        for key in ["visit_date", "exam_time"]:
            if key in data.keys() and data[key] is not None:
                data[key] = datetime.datetime.fromisoformat(data[key])
        return cls(**data)



from collections import defaultdict
import json
import logging
from pathlib import Path
import shutil
import tempfile
from typing import (Callable, List, Optional, overload, SupportsIndex, Tuple,
                    TypedDict, Union)
import warnings
import zipfile

from matplotlib import patches
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from numpy import typing as npt
import numpy as np
from skimage import transform
from skimage.transform._geometric import GeometricTransform

from eyepy import config
from eyepy.core.annotations import EyeVolumeLayerAnnotation
from eyepy.core.annotations import EyeVolumePixelAnnotation
from eyepy.core.eyebscan import EyeBscan
from eyepy.core.eyeenface import EyeEnface
from eyepy.core.eyemeta import EyeBscanMeta
from eyepy.core.eyemeta import EyeEnfaceMeta
from eyepy.core.eyemeta import EyeVolumeMeta
from eyepy.core.utils import intensity_transforms

logger = logging.getLogger("eyepy.core.eyevolume")


class EyeVolume:
    """ """

    def __init__(
        self,
        data: npt.NDArray[np.float32],
        meta: Optional[EyeVolumeMeta] = None,
        localizer: Optional[EyeEnface] = None,
        transformation: Optional[GeometricTransform] = None,
    ):
        """

        Args:
            data: A 3D numpy array containing the OCT data in shape (n_bscans, bscan_height, bscan_width)
            meta: Optional [EyeVolumeMeta][eyepy.core.eyemeta.EyeVolumeMeta] object.
            localizer:
            transformation:
        """
        self._raw_data = data
        self._data = None

        self._bscans = {}

        if meta is None:
            self.meta = self._default_meta(self._raw_data)
        else:
            self.meta = meta
        if not "intensity_transform" in self.meta:
            self.meta["intensity_transform"] = "default"

        self.set_intensity_transform(self.meta["intensity_transform"])

        self._layers = []
        self._volume_maps = []
        self._ascan_maps = []

        if transformation is None:
            self.localizer_transform = self._estimate_transform()
        else:
            self.localizer_transform = transformation

        if localizer is None:
            self.localizer = self._default_localizer(self.data)
        else:
            self.localizer = localizer

    def save(self, path):
        """

        Args:
            path:

        Returns:

        """
        # Create temporary folder
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpdirname = Path(tmpdirname)

            # Save OCT volume as npy and meta as json
            np.save(tmpdirname / "raw_volume.npy", self._raw_data)
            with open(tmpdirname / "meta.json", "w") as meta_file:
                if self.meta["intensity_transform"] == "custom":
                    warnings.warn(
                        "Custom intensity transforms can not be saved.")
                    self.meta["intensity_transform"] = "default"
                json.dump(self.meta.as_dict(), meta_file)

            if not len(self._volume_maps) == 0:
                # Save Volume Annotations
                voxels_path = tmpdirname / "annotations" / "voxels"
                voxels_path.mkdir(exist_ok=True, parents=True)
                np.save(
                    voxels_path / "voxel_maps.npy",
                    np.stack([v.data for v in self._volume_maps]),
                )
                with open(voxels_path / "meta.json", "w") as meta_file:
                    json.dump([v.meta for v in self._volume_maps], meta_file)

            if not len(self._layers) == 0:
                # Save layer annotations
                layers_path = tmpdirname / "annotations" / "layers"
                layers_path.mkdir(exist_ok=True, parents=True)
                np.save(
                    layers_path / "layer_heights.npy",
                    np.stack([l.data for l in self._layers]),
                )
                with open(layers_path / "meta.json", "w") as meta_file:
                    json.dump([l.meta for l in self._layers], meta_file)

            # Save Localizer
            localizer_path = tmpdirname / "localizer"
            localizer_path.mkdir(exist_ok=True, parents=True)
            np.save(localizer_path / "localizer.npy", self.localizer.data)
            with open(localizer_path / "meta.json", "w") as meta_file:
                json.dump(self.localizer.meta.as_dict(), meta_file)

            # Save Localizer Annotations
            if not len(self.localizer._area_maps) == 0:
                pixels_path = localizer_path / "annotations" / "pixel"
                pixels_path.mkdir(exist_ok=True, parents=True)
                np.save(
                    pixels_path / "pixel_maps.npy",
                    np.stack([p.data for p in self.localizer._area_maps]),
                )
                with open(pixels_path / "meta.json", "w") as meta_file:
                    json.dump([p.meta for p in self.localizer._area_maps],
                              meta_file)

            # Zip and copy to location
            name = shutil.make_archive(str(tmpdirname / Path(path).stem),
                                       "zip", tmpdirname)
            shutil.move(name, path)

    @classmethod
    def load(cls, path):
        """

        Args:
            path:

        Returns:

        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpdirname = Path(tmpdirname)
            with zipfile.ZipFile(path, "r") as zip_ref:
                zip_ref.extractall(tmpdirname)

            # Load raw volume and meta
            data = np.load(tmpdirname / "raw_volume.npy")
            with open(tmpdirname / "meta.json", "r") as meta_file:
                volume_meta = EyeVolumeMeta.from_dict(json.load(meta_file))

            # Load Volume Annotations
            voxels_path = tmpdirname / "annotations" / "voxels"
            if voxels_path.exists():
                voxel_annotations = np.load(voxels_path / "voxel_maps.npy")
                with open(voxels_path / "meta.json", "r") as meta_file:
                    voxels_meta = json.load(meta_file)
            else:
                voxel_annotations = []
                voxels_meta = []

            # Load layers
            layers_path = tmpdirname / "annotations" / "layers"
            if layers_path.exists():
                layer_annotations = np.load(layers_path / "layer_heights.npy")
                with open(layers_path / "meta.json", "r") as meta_file:
                    layers_meta = json.load(meta_file)

                # Clean knots
                for i, layer_meta in enumerate(layers_meta):
                    if "knots" in layer_meta:
                        knots = layer_meta["knots"]
                        knots = {int(i): knots[i] for i in knots}
                        layer_meta["knots"] = knots
            else:
                layer_annotations = []
                layers_meta = []

            # Load Localizer and meta
            localizer_path = tmpdirname / "localizer"
            localizer_data = np.load(localizer_path / "localizer.npy")
            with open(localizer_path / "meta.json", "r") as meta_file:
                localizer_meta = EyeEnfaceMeta.from_dict(json.load(meta_file))
            localizer = EyeEnface(data=localizer_data, meta=localizer_meta)

            # Load Localizer Annotations
            pixels_path = localizer_path / "annotations" / "pixel"
            if pixels_path.exists():
                pixel_annotations = np.load(pixels_path / "pixel_maps.npy")
                with open(pixels_path / "meta.json", "r") as meta_file:
                    pixels_meta = json.load(meta_file)

                for i, pixel_meta in enumerate(pixels_meta):
                    localizer.add_area_annotation(pixel_annotations[i],
                                                  pixel_meta)

            from eyepy.io.utils import _compute_localizer_oct_transform

            transformation = _compute_localizer_oct_transform(
                volume_meta, localizer_meta, data.shape)

            ev = cls(
                data=data,
                meta=volume_meta,
                localizer=localizer,
                transformation=transformation,
            )
            for meta, annotation in zip(layers_meta, layer_annotations):
                ev.add_layer_annotation(annotation, meta)

            for meta, annotation in zip(voxels_meta, voxel_annotations):
                ev.add_pixel_annotation(annotation, meta)

        return ev

    def _default_meta(self, volume):
        bscan_meta = [
            EyeBscanMeta(start_pos=(0, i),
                         end_pos=((volume.shape[2] - 1), i),
                         pos_unit="pixel")
            for i in range(volume.shape[0] - 1, -1, -1)
        ]
        meta = EyeVolumeMeta(
            scale_x=1.0,
            scale_y=1.0,
            scale_z=1.0,
            scale_unit="pixel",
            intensity_transform="default",
            bscan_meta=bscan_meta,
        )
        return meta

    def _default_localizer(self, data):
        projection = np.flip(np.nanmean(data, axis=1), axis=0)
        image = transform.warp(
            projection,
            self.localizer_transform.inverse,
            output_shape=(self.size_x, self.size_x),
            order=1,
        )
        localizer = EyeEnface(
            image,
            meta=EyeEnfaceMeta(
                scale_x=self.scale_x,
                scale_y=self.scale_x,
                scale_unit=self.scale_unit,
                field_size=0,
                scan_focus=0,
                laterality="NA",
            ),
        )
        return localizer

    def _estimate_transform(self):
        # Compute a transform to map a 2D projection of the volume to a square
        # Points in oct space
        src = np.array([
            [0, 0],  # Top left
            [0, self.size_x - 1],  # Top right
            [self.size_z - 1, 0],  # Bottom left
            [self.size_z - 1, self.size_x - 1],
        ])  # Bottom right

        # Respective points in enface space
        dst = np.array([
            (0, 0),  # Top left
            (0, self.size_x - 1),  # Top right
            (self.size_x - 1, 0),  # Bottom left
            (self.size_x - 1, self.size_x - 1),
        ])  # Bottom right

        # Switch from x/y coordinates to row/column coordinates for src and dst
        src = np.flip(src, axis=1)
        dst = np.flip(dst, axis=1)
        # src = src[:, [1, 0]]
        # dst = dst[:, [1, 0]]

        return transform.estimate_transform("affine", src, dst)

    @overload
    def __getitem__(self, index: SupportsIndex) -> EyeBscan:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[EyeBscan]:
        ...

    def __getitem__(
            self, index: Union[SupportsIndex,
                               slice]) -> Union[List[EyeBscan], EyeBscan]:
        """

        Args:
            index:

        Returns:

        """
        # The B-Scan at the given index.
        if isinstance(index, slice):
            return [self[i] for i in range(*index.indices(len(self)))]
        elif isinstance(index, int):
            if index < 0:
                index = len(self) + index

            if index < len(self):
                try:
                    # Return B-scan with type annotation
                    return self._bscans[index]
                except KeyError:
                    self._bscans[index] = EyeBscan(self, index)
                    return self._bscans[index]
            else:
                raise IndexError()
        else:
            raise TypeError()

    def __len__(self):
        """The number of B-Scans."""
        return self.shape[0]

    def set_intensity_transform(self, func: Union[str, Callable]):
        """

        Args:
            func: Either a string specifying a transform from eyepy.core.utils.intensity_transforms or a function

        Returns:

        """
        if isinstance(func, str):
            if func in intensity_transforms:
                self.meta["intensity_transform"] = func
                self.intensity_transform = intensity_transforms[func]
                self._data = None
            elif func == "custom":
                logger.warning(
                    "Custom intensity transforms can not be loaded currently")
            else:
                logger.warning(
                    f"Provided intensity transform name {func} is not known. Valid names are 'vol' or 'default'. You can also pass your own function."
                )
        elif isinstance(func, Callable):
            self.meta["intensity_transform"] = "custom"
            self.intensity_transform = func
            self._data = None

    @property
    def data(self):
        """

        Returns:

        """
        if self._data is None:
            self._data = self.intensity_transform(np.copy(self._raw_data))
        return self._data

    @property
    def shape(self):
        """

        Returns:

        """
        return self._raw_data.shape

    @shape.setter
    def shape(self, value):
        raise AttributeError(
            "Shape can not be set since it is derived from the data")

    @property
    def scale(self):
        """

        Returns:

        """
        return self.scale_z, self.scale_y, self.scale_x

    @scale.setter
    def scale(self, value):
        self.scale_z, self.scale_y, self.scale_x = value

    @property
    def size_z(self):
        """

        Returns:

        """
        return self.shape[0]

    @size_z.setter
    def size_z(self, value):
        raise AttributeError(
            "Size of z axis can not be changed since it is derived from the data"
        )

    @property
    def size_y(self):
        """

        Returns:

        """
        return self.shape[1]

    @size_y.setter
    def size_y(self, value):
        raise AttributeError(
            "Size of y axis can not be changed since it is derived from the data"
        )

    @property
    def size_x(self):
        """

        Returns:

        """
        return self.shape[2]

    @size_x.setter
    def size_x(self, value):
        raise AttributeError(
            "Size of x axis can not be changed since it is derived from the data"
        )

    @property
    def scale_z(self):
        """

        Returns:

        """
        return self.meta["scale_z"]

    @scale_z.setter
    def scale_z(self, value):
        self.meta["scale_z"] = value

    @property
    def scale_y(self):
        """

        Returns:

        """
        return self.meta["scale_y"]

    @scale_y.setter
    def scale_y(self, value):
        self.meta["scale_y"] = value

    @property
    def scale_x(self):
        """

        Returns:

        """
        return self.meta["scale_x"]

    @scale_x.setter
    def scale_x(self, value):
        self.meta["scale_x"] = value

    @property
    def scale_unit(self):
        """

        Returns:

        """
        return self.meta["scale_unit"]

    @scale_unit.setter
    def scale_unit(self, value):
        self.meta["scale_unit"] = value

    @property
    def laterality(self):
        """

        Returns:

        """
        return self.meta["laterality"]

    @laterality.setter
    def laterality(self, value):
        self.meta["laterality"] = value

    @property
    def layers(self):
        """

        Returns:

        """
        # Create a dict to access layers by their name. Keys in this dict always
        # reflect the current name attribute of the layers.
        return {layer.name: layer for layer in self._layers}

    @property
    def volume_maps(self):
        """

        Returns:

        """
        # Create a dict to access volume_maps by their name
        return {vm.name: vm for vm in self._volume_maps}

    def add_pixel_annotation(self, voxel_map=None, meta=None, **kwargs):
        """

        Args:
            voxel_map:
            meta:
            **kwargs:

        Returns:

        """
        if meta is None:
            meta = {}
        meta.update(**kwargs)
        voxel_annotation = EyeVolumePixelAnnotation(self, voxel_map, **meta)
        self._volume_maps.append(voxel_annotation)
        return voxel_annotation

    def remove_pixel_annotations(self, name):
        """

        Args:
            name:

        Returns:

        """
        for i, voxel_map in enumerate(self._volume_maps):
            if voxel_map.name == name:
                self._volume_maps.pop(i)

        # Remove references from B-scans
        for bscan in self:
            if name in bscan.area_maps:
                bscan.area_maps.pop(name)

    def add_layer_annotation(self, height_map=None, meta=None, **kwargs):
        """

        Args:
            height_map: Height in shape (n_Bscans, Bscan_width) The first index refers to the bottom most B-scan
            meta: name, current_color, and knots
            **kwargs:

        Returns:

        """
        if meta is None:
            meta = {}
        meta.update(**kwargs)
        layer_annotation = EyeVolumeLayerAnnotation(self, height_map, **meta)
        self._layers.append(layer_annotation)
        return layer_annotation

    def remove_layer_annotation(self, name):
        """

        Args:
            name:

        Returns:

        """
        for i, layer in enumerate(self._layers):
            if layer.name == name:
                self._layers.pop(i)

        # Remove references from B-scans
        for bscan in self:
            if name in bscan.layers:
                bscan.layers.pop(name)

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        projections: Union[bool, List[str]] = False,
        bscan_region: bool = False,
        bscan_positions: Union[bool, List[int]] = False,
        quantification: Optional[str] = None,
        region: Union[slice, Tuple[slice, slice]] = np.s_[:, :],
        annotations_only: bool = False,
        projection_kwargs: Optional[dict] = None,
        line_kwargs: Optional[dict] = None,
    ) -> None:
        """ Plot an annotated OCT localizer image. If the volume does not provide a localizer image an enface projection of the OCT volume is used instead.

        Args:
            ax: Axes to plot on. If not provided plot on the current axes (plt.gca()).
            projections: If `True` plot all projections (default: `False`). If a list of strings is given, plot the projections with the given names. Projections are 2D enface views on oct volume annotations such as drusen.
            bscan_region: If `True` plot the region B-scans are located in (default: `False`)
            bscan_positions: If `True` plot positions of all B-scan (default: `False`). If a list of integers is given, plot the B-scans with the respective indices. Indexing starts at the bottom of the localizer.
            quantification: Name of the OCT volume annotations to plot a quantification for (default: `None`). Quantifications are performed on circular grids.
            region: Region of the localizer to plot (default: `np.s_[...]`)
            annotations_only: If `True` localizer image is not plotted (defaualt: `False`)
            projection_kwargs: Optional keyword arguments for the projection plots. If `None` default values are used (default: `None`). If a dictionary is given, the keys are the projection names and the values are dictionaries of keyword arguments.
            line_kwargs: Optional keyword arguments for customizing the lines to show B-scan region and positions plots. If `None` default values are used which are {"linewidth": 0.2, "linestyle": "-", "color": "green"}

        Returns:
            None

        """

        # Complete region index expression
        y_start = region[0].start if region[0].start is not None else 0
        y_stop = region[0].stop if region[
            0].stop is not None else self.localizer.shape[0]
        x_start = region[1].start if region[1].start is not None else 0
        x_stop = region[1].stop if region[
            1].stop is not None else self.localizer.shape[1]

        region = np.s_[y_start:y_stop, x_start:x_stop]

        if ax is None:
            ax = plt.gca()

        if not annotations_only:
            self.localizer.plot(ax=ax, region=region)

        if projections is True:
            projections = list(self.volume_maps.keys())
        elif not projections:
            projections = []

        if projection_kwargs is None:
            projection_kwargs = defaultdict(lambda: {})
        for name in projections:
            if not name in projection_kwargs.keys():
                projection_kwargs[name] = {}
            self.volume_maps[name].plot(ax=ax,
                                        region=region,
                                        **projection_kwargs[name])

        if line_kwargs is None:
            line_kwargs = config.line_kwargs
        else:
            line_kwargs = {**config.line_kwargs, **line_kwargs}

        if bscan_positions:
            self._plot_bscan_positions(
                ax=ax,
                bscan_positions=bscan_positions,
                region=region,
                line_kwargs=line_kwargs,
            )
        if bscan_region:
            self._plot_bscan_region(region=region,
                                    ax=ax,
                                    line_kwargs=line_kwargs)

        if quantification:
            self.volume_maps[quantification].plot_quantification(region=region,
                                                                 ax=ax)

    def _plot_bscan_positions(
        self,
        bscan_positions: Union[bool, List[int]] = True,
        ax=None,
        region: Union[slice, Tuple[slice, slice]] = np.s_[:, :],
        line_kwargs=None,
    ):
        if not bscan_positions:
            bscan_positions = []
        elif bscan_positions is True:
            bscan_positions = list(range(0, len(self)))

        if ax is None:
            ax = plt.gca()
        if line_kwargs is None:
            line_kwargs = {}

        for i in bscan_positions:
            scale = np.array([self.localizer.scale_x, self.localizer.scale_y])

            start = self[i].meta["start_pos"] / scale
            end = self[i].meta["end_pos"] / scale

            for pos in [start, end]:
                # Check for both axis if pos is in region
                if not (region[0].start <= pos[0] <= region[0].stop
                        and region[1].start <= pos[1] <= region[1].stop):
                    logger.warning(
                        "B-scan position can not be plotted because the visualized region does not contain the complete B-scan."
                    )
                    return

            polygon = patches.Polygon(
                [start, end],
                closed=False,
                fill=False,
                alpha=1,
                antialiased=False,
                rasterized=False,
                snap=False,
                **line_kwargs,
            )
            ax.add_patch(polygon)

    def _plot_bscan_region(self,
                           region: Union[slice, Tuple[slice,
                                                      slice]] = np.s_[:, :],
                           ax=None,
                           line_kwargs=None):
        if ax is None:
            ax = plt.gca()

        if line_kwargs is None:
            line_kwargs = {}

        scale = np.array([self.localizer.scale_x, self.localizer.scale_y])

        upper_left = self[-1].meta["start_pos"] / scale
        lower_left = self[0].meta["start_pos"] / scale
        lower_right = self[0].meta["end_pos"] / scale
        upper_right = self[-1].meta["end_pos"] / scale

        for pos in [upper_left, lower_left, lower_right, upper_right]:
            # Check for both axis if pos is in region
            if not (region[0].start <= pos[0] <= region[0].stop
                    and region[1].start <= pos[1] <= region[1].stop):
                logger.warning(
                    "B-scan region can not be plotted because the visualized region does not contain the complete B-scan region."
                )
                return

        polygon = patches.Polygon(
            [upper_left, lower_left, lower_right, upper_right],
            closed=True,
            fill=False,
            alpha=1,
            antialiased=False,
            rasterized=False,
            snap=False,
            **line_kwargs,
        )
        ax.add_patch(polygon)



import logging
from pathlib import Path
from typing import List, Union

import imageio.v2 as imageio
import numpy as np

from eyepy import EyeBscanMeta
from eyepy import EyeVolume
from eyepy import EyeVolumeMeta

from .he import HeE2eReader
from .he import HeVolReader
from .he import HeVolWriter
from .he import HeXmlReader

logger = logging.getLogger("eyepy.io")


def import_heyex_e2e(path: Union[str, Path]) -> EyeVolume:
    """ Read a Heyex E2E file

    This function is a thin wrapper around the HeE2eReader class and
    returns the first of potentially multiple OCT volumes. If you want to
    read all volumes, or need more control, you can use the
    [HeE2eReader][eyepy.io.he.e2e_reader.HeE2eReader] class directly.

    Args:
        path: Path to the E2E file

    Returns:
        Parsed data as EyeVolume object

    """
    reader = HeE2eReader(path)
    if len(reader.series) < 1:
        logger.info(
            f"There are {len(reader.series)} Series stored in the E2E file. If you want to read all of them, use the HeE2eReader class directly."
        )
    with reader as open_reader:
        ev = open_reader.volume
    return ev


def import_heyex_xml(path: Union[str, Path]) -> EyeVolume:
    """ Read a Heyex XML file

    This function is a thin wrapper around the HeXmlReader class
    which you can use directly if you need more control.

    Args:
        path: Path to the XML file or the folder containing the XML file

    Returns:
        Parsed data as EyeVolume object

    """
    return HeXmlReader(path).volume


def import_heyex_vol(path: Union[str, Path]) -> EyeVolume:
    """ Read a Heyex VOL file

    This function is a thin wrapper around the HeVolReader class
    which you can use directly if you need more control.

    Args:
        path: Path to the VOL file

    Returns:
        Parsed data as EyeVolume object

    """
    return HeVolReader(path).volume


def import_bscan_folder(path: Union[str, Path]) -> EyeVolume:
    """ Read B-Scans from a folder

    This function can be used to read B-scans from a folder in case that
    there is no additional metadata available.

    Args:
        path: Path to the folder containing the B-Scans

    Returns:
        Parsed data as EyeVolume object

    """
    path = Path(path)
    img_paths = sorted(list(path.iterdir()))
    img_paths = [
        p for p in img_paths if p.is_file()
        and p.suffix.lower() in [".jpg", ".jpeg", ".tiff", ".tif", ".png"]
    ]

    images = []
    for p in img_paths:
        image = imageio.imread(p)
        if len(image.shape) == 3:
            image = image[..., 0]
        images.append(image)

    volume = np.stack(images, axis=0)
    bscan_meta = [
        EyeBscanMeta(start_pos=(0, i),
                     end_pos=(volume.shape[2] - 1, i),
                     pos_unit="pixel")
        for i in range(volume.shape[0] - 1, -1, -1)
    ]
    meta = EyeVolumeMeta(scale_x=1,
                         scale_y=1,
                         scale_z=1,
                         scale_unit="pixel",
                         bscan_meta=bscan_meta)

    return EyeVolume(data=volume, meta=meta)


def import_duke_mat(path: Union[str, Path]) -> EyeVolume:
    """ Import an OCT volume from the Duke dataset

    The dataset is available at https://people.duke.edu/~sf59/RPEDC_Ophth_2013_dataset.htm
    OCT volumes are stored as .mat files which are parsed by this function and returned as
    EyeVolume object.

    Args:
        path: Path to the .mat file

    Returns:
        Parsed data as EyeVolume object

    """
    import scipy.io as sio

    loaded = sio.loadmat(path)
    volume = np.moveaxis(loaded["images"], -1, 0)
    layer_maps = np.moveaxis(loaded["layerMaps"], -1, 0)

    bscan_meta = [
        EyeBscanMeta(
            start_pos=(0, 0.067 * i),
            end_pos=(0.0067 * (volume.shape[2] - 1), 0.067 * i),
            pos_unit="mm",
        ) for i in range(volume.shape[0] - 1, -1, -1)
    ]
    meta = EyeVolumeMeta(
        scale_x=0.0067,
        scale_y=0.0045,  # https://retinatoday.com/articles/2008-may/0508_10-php
        scale_z=0.067,
        scale_unit="mm",
        bscan_meta=bscan_meta,
        age=loaded["Age"],
    )

    volume = EyeVolume(data=volume, meta=meta)
    names = {0: "ILM", 1: "IBRPE", 2: "BM"}
    for i, height_map in enumerate(layer_maps):
        volume.add_layer_annotation(
            np.flip(height_map, axis=0),
            name=names[i],
        )

    return volume


def import_retouch(path: Union[str, Path]) -> EyeVolume:
    """ Import an OCT volume from the Retouch dataset

    The dataset is available upon request at https://retouch.grand-challenge.org/
    Reading the data requires the ITK library to be installed. You can install it with pip:

    `pip install itk`

    Args:
        path: Path to the folder containing the OCT volume

    Returns:
        Parsed data as EyeVolume object

    """
    import itk

    path = Path(path)
    data = itk.imread(str(path / "oct.mhd"))

    bscan_meta = [
        EyeBscanMeta(
            start_pos=(0, data["spacing"][0] * i),
            end_pos=(data["spacing"][2] * (data.shape[2] - 1),
                     data["spacing"][0] * i),
            pos_unit="mm",
        ) for i in range(data.shape[0] - 1, -1, -1)
    ]

    meta = EyeVolumeMeta(
        scale_x=data["spacing"][2],
        scale_y=data["spacing"][1],
        scale_z=data["spacing"][0],
        scale_unit="mm",
        bscan_meta=bscan_meta,
    )
    # Todo: Add intensity transform instead. Topcon and Cirrus are stored as UCHAR while heidelberg is stored as USHORT
    data = (data[...].astype(float) / np.iinfo(data[...].dtype).max *
            255).astype(np.uint8)
    eye_volume = EyeVolume(data=data[...], meta=meta)

    if (path / "reference.mhd").is_file():
        annotation = itk.imread(str(path / "reference.mhd"))
        eye_volume.add_pixel_annotation(np.equal(annotation, 1),
                                        name="IRF",
                                        current_color="FF0000")
        eye_volume.add_pixel_annotation(np.equal(annotation, 2),
                                        name="SRF",
                                        current_color="0000FF")
        eye_volume.add_pixel_annotation(np.equal(annotation, 3),
                                        name="PED",
                                        current_color="FFFF00")

    return eye_volume



from typing import Tuple

import numpy as np
import numpy.typing as npt
from skimage import img_as_float32
from skimage import img_as_ubyte

from eyepy.core.filter import filter_by_height_enface

from .annotations import EyeVolumeLayerAnnotation

NDArrayFloat = npt.NDArray[np.float_]
NDArrayBool = npt.NDArray[np.bool_]


class DynamicDefaultDict(dict):
    """A defaultdict for which the factory function has access to the missing key"""

    def __init__(self, factory):
        self.factory = factory

    def __missing__(self, key):
        self[key] = self.factory(key)
        return self[key]


def vol_intensity_transform(data: np.ndarray) -> np.ndarray:
    """ Wrapper around from_vol_intensity

    Transform intensities from Heyex VOL exports to achieve a constrast similar to the one used in Heyex.

    Args:
        data: Input data

    Returns:
        Transformed data
    """
    return from_vol_intensity(data)


def from_vol_intensity(data):
    selection_0 = data == np.finfo(np.float32).max
    selection_data = data <= 1

    new = np.log(data[selection_data] + 2.44e-04)
    new = (new + 8.3) / 8.285

    data[selection_data] = new
    data[selection_0] = 0
    data = np.clip(data, 0, 1)
    return img_as_ubyte(data)


# Function expects numpy array of uint8 type hint
def to_vol_intensity(data):
    data = img_as_float32(data)
    data = data * 8.285 - 8.3
    data = np.exp(data) - 2.44e-04
    return data


def default_intensity_transform(data: np.ndarray) -> np.ndarray:
    """ Default intensity transform

    By default intensities are not changed.

    Args:
        data: Input data

    Returns:
        Input data unchanged
    """
    return data


intensity_transforms = {
    "default": default_intensity_transform,
    "vol": vol_intensity_transform,
}


def ideal_rpe(rpe_height: NDArrayFloat, bm_height: NDArrayFloat,
              volume_shape: Tuple[int, int, int]) -> NDArrayFloat:
    """ Compute the ideal RPE from an RPE with Drusen.

    Args:
        rpe_height: The RPE height as offset from the lower border of the B-Scan
        bm_height: The BM height as offset from the lower border of the B-Scan
        volume_shape: Shape of the OCT volume (number of B-Scans, height, width)

    Returns:
        The ideal RPE height as offset from the lower border of the B-Scan
    """
    d, h, w = volume_shape

    # compute shift needed to align the BM to the horizontal center line
    shift = np.empty((d, w), dtype="int")
    shift.fill(h - (h / 2))
    shift = shift - bm_height

    # now shift the RPE location array as well
    shifted_rpe_height = rpe_height + shift

    # Remove all NANs from the shifted RPE data
    clean_shifted = shifted_rpe_height[~np.isnan(shifted_rpe_height)]

    # Compute a histogram with a bin for every pixel height in a B-Scan
    hist, edges = np.histogram(clean_shifted.flatten(),
                               bins=np.arange(volume_shape[1]))

    # Compute the ideal RPE as the mean of the biggest bin and its neighbours
    lower_edge = edges[np.argmax(hist) - 1]
    upper_edge = edges[np.argmax(hist) + 2]
    irpe_height = np.mean(clean_shifted[np.logical_and(
        clean_shifted <= upper_edge, clean_shifted >= lower_edge)])
    ideal_rpe = np.full_like(shifted_rpe_height, irpe_height)

    # Shift back into original image space
    ideal_rpe = np.reshape(ideal_rpe, (d, w)) - shift

    return ideal_rpe


def drusen(rpe_height: NDArrayFloat,
           bm_height: NDArrayFloat,
           volume_shape: Tuple[int, int, int],
           minimum_height: int = 2) -> NDArrayBool:
    """Compute drusen from the RPE and BM layer segmentation.

    First estimate the ideal RPE based on a histogram of the RPE heights relativ
    to the BM. Then compute drusen as the area between the RPE and the normal RPE

    Args:
        rpe_height: The RPE height as offset from the lower border of the B-Scan
        bm_height: The BM height as offset from the lower border of the B-Scan
        volume_shape: Shape of the OCT volume (number of B-Scans, height, width)
        minimum_height: Minimum height of a drusen in pixels

    Returns:
        A boolean array with the same shape as the OCT volume. True indicates a
        voxel beeing part of a drusen.
    """
    # Estimate ideal RPE
    if isinstance(rpe_height, EyeVolumeLayerAnnotation):
        rpe_height = np.copy(rpe_height.data)
    if isinstance(bm_height, EyeVolumeLayerAnnotation):
        bm_height = np.copy(bm_height.data)

    irpe = ideal_rpe(rpe_height, bm_height, volume_shape)
    # Create drusen map
    drusen_map = np.zeros(volume_shape, dtype=bool)
    # Exclude normal RPE and RPE from the drusen area.
    nans = np.isnan(rpe_height + irpe)

    rpe = np.rint(rpe_height + 1)
    rpe[nans] = 0
    rpe = rpe.astype(int)

    irpe = np.rint(irpe)
    irpe[nans] = 0
    irpe = irpe.astype(int)

    for sli in range(drusen_map.shape[0]):
        for col in range(drusen_map.shape[2]):
            if not nans[sli, col]:
                drusen_map[sli, rpe[sli, col]:irpe[sli, col], col] = 1

    drusen_map = filter_by_height_enface(drusen_map, minimum_height)

    return drusen_map
